﻿Option Strict Off
Option Explicit On
Imports MSCommLib
Imports System.IO.Ports
Imports System.Text.RegularExpressions

Public Class Main
    Dim WithEvents MSComm1 As New MSComm
    Dim Inputstr As String = ""     'Sammelstring für Receive-Bytes
    Dim connected As Boolean
    Dim initialized As Boolean = False  'Um zu verhindern, dass Changed-Events von Elementen (z.B. Radiobuttons, NumericUpDowns, etc.) während der Initialisierung ausgeführt werden
    Dim SetMark As Boolean          'Flag, dass in Trace-Grafiken ein Marker gesetzt werden soll
    Dim Tab1_TB() As TextBox        'Das Array der Textboxen in Tab1
    Dim Tab1_PB() As PictureBox     'Das Array der Pictureboxen in Tab1
    Dim Tab1_LB() As Label          'Das Array der Label in Tab1
    Dim Tab1_SID As Integer = &H20  'Die aktuell in Tab1 getracte SID (Startwert: 020)

    Dim Tab4_ND() As NumericUpDown  'Das Array der Data-Felder in Tab4
    Dim Tab4_LD() As Label          'Das Array der Label in Tab4

    Private Sub Main_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        Dim ports As String() = SerialPort.GetPortNames()

        If ports.Length > 0 Then
            With MSComm1
                .CommPort = Val(Mid(ports(0), 4))
                .Settings = "9600,N,8,1"
                .RThreshold = 1             'comEvReceive-Event wird ausgeloest wenn 1 Byte angekommen ist. Jedes angekommene Byte soll sofort protokolliert werden
                .InputLen = 0               'Input liest alle Bytes im InputBuffer
                .PortOpen = True
            End With
            L_SimRcv.Visible = False
            T_SimRcv.Visible = False
            B_SimRcv.Visible = False
            Me.Text = "CAN-FD Sniffer - an " + ports(0)
            connected = True
        Else
            L_SimRcv.Visible = True
            T_SimRcv.Visible = True
            B_SimRcv.Visible = True
            Me.Text = "CAN-FD Sniffer - Simulationsmodus"
            connected = False
        End If

        Tab0_Init()
        Tab1_Init()
        Tab2_Init()
        Tab4_Init()
        SetMark = False
        initialized = True
    End Sub

    Private Sub Main_FormClosed(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosedEventArgs) Handles Me.FormClosed
        If connected = True Then MSComm1.PortOpen = False
        Application.Exit()
    End Sub

    Private Sub MSComm1_OnComm() Handles MSComm1.OnComm
        Select Case MSComm1.CommEvent
            Case MSCommLib.CommEventConstants.comEventOverrun
                MSComm1.PortOpen = False
                MsgBox("comEventOverrun aufgetreten")
            Case MSCommLib.CommEventConstants.comEventRxOver
                MSComm1.PortOpen = False
                MsgBox("comEventRxOver aufgetreten")
            Case MSCommLib.OnCommConstants.comEvReceive
                Dim input As String
                Dim ch As String
                Dim int As Integer
                Dim n As Integer
                input = MSComm1.Input       'Daten abholen
                Inputstr = Inputstr + input     'In InputStr sammeln - wird in WaitForUARTReturn ausgewertet
                If Ch_RxProt.Checked = True Then
                    For n = 1 To Len(input)         'Empfang sofort protokollieren. Sollte nur 1 Byte sein, trotzdem Schleife, falls es doch mal mehr sein sollten
                        ch = Mid(input, n, 1)
                        int = Asc(ch)
                        T_RX.Text = T_RX.Text + int.ToString("X2") + ","
                    Next
                End If
        End Select
    End Sub

    Private Sub B_SimRcv_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles B_SimRcv.Click
        'Simuliert UART-Empfang
        Dim datstr, valtxt, valtxtmod As String
        Dim valtxtarray As String()
        Dim val, n As Integer
        Dim ch As String
        Dim int As Integer

        'Aus den Eingaben der Textbox einen Binary-String erstellen
        datstr = ""
        valtxtarray = (T_SimRcv.Text).Split(",")
        For Each valtxt In valtxtarray
            valtxtmod = ""
            For n = 1 To valtxt.Length                                      'Alle Zeichen, die nicht Hex sind entfernen
                If Regex.IsMatch(Mid(valtxt, n, 1), "[a-fA-F0-9]") Then
                    valtxtmod = valtxtmod + Mid(valtxt, n, 1)
                End If
            Next
            If valtxtmod.Length > 0 Then
                If valtxtmod.Length > 2 Then                          'Bei mehr als zwei Stellen, nur die letzten 2 Stellen weiterverwenden
                    valtxtmod = Mid(valtxtmod, valtxtmod.Length - 1, 2)
                End If
                val = CInt("&H" + valtxtmod)
                datstr = datstr + Chr(val)
            End If
        Next
        'mit DatenString verfahren, wie es auch der MSComm-Receive-Event tun würde 
        Inputstr = Inputstr + datstr     'In InputStr sammeln - wird in WaitForUARTReturn ausgewertet
        If Ch_RxProt.Checked = True Then
            For n = 1 To Len(datstr)        'Receive protokollieren
                ch = Mid(datstr, n, 1)
                int = Asc(ch)
                T_RX.Text = T_RX.Text + int.ToString("X2") + ","
            Next
        End If
    End Sub

    Private Sub B_TXclear_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles B_TXclear.Click
        T_TX.Clear()
    End Sub

    Private Sub B_RXclear_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles B_RXclear.Click
        T_RX.Clear()
    End Sub

    Private Sub UART_ClearInput()
        Inputstr = ""
    End Sub

    Private Sub UARTSend(ByVal outstr As String)
        Dim ch As String
        Dim int As Integer
        Dim n As Integer

        If connected = True Then MSComm1.Output = outstr
        If Ch_TxProt.Checked = True Then
            For n = 1 To Len(outstr)         'Sendung sofort protokollieren.
                ch = Mid(outstr, n, 1)
                int = Asc(ch)
                T_TX.Text = T_TX.Text + int.ToString("X2") + ","
            Next
            T_TX.Text = T_TX.Text + vbCrLf
        End If
    End Sub

    Private Function WaitForUARTReturn(ByVal anz As Integer) As String
        'Alt aber proven
        'Inputstr = ""
        'If connected = False Then
        '    MsgBox("Es werden " + anz.ToString + "Bytes Empfang erwartet. Bytes über Sim-Receive senden!")
        '    Dim n As Integer
        '    Dim txt As String = ""
        '    For n = 1 To anz
        '        txt = txt + ",00"
        '    Next
        '    txt = Mid(txt, 2)
        '    T_SimRcv.Text = txt
        'End If
        'While Inputstr.Length < anz                     'Warte bis anz Bytes empfangen wurden
        '    System.Windows.Forms.Application.DoEvents()
        'End While
        'Return Mid(Inputstr, 1, anz)                    'Anz Byte zurückliefern

        Dim ret As String

        If connected = False Then
            MsgBox("Es werden " + anz.ToString + "Bytes Empfang erwartet. Bytes über Sim-Receive senden!")
            Dim n As Integer
            Dim txt As String = ""
            For n = 1 To anz
                txt = txt + ",00"
            Next
            txt = Mid(txt, 2)
            T_SimRcv.Text = txt
        End If
        While Inputstr.Length < anz                     'Warte bis anz Bytes empfangen wurden
            System.Windows.Forms.Application.DoEvents()
        End While
        ret = Mid(Inputstr, 1, anz)                     'Die ersten anz Byte sollen zurückgeliefert werden
        Inputstr = Mid(Inputstr, anz + 1)                   'Der Rest soll im Inputstr verbleiben
        Return ret                                      'Anz Byte zurückliefern
    End Function


    'Funktionen des Remote-Tab
    Private Sub Tab0_Init()
        'Combo-Boxen für Baudraten füllen

        Tab0_Cbsbaud.Items.Add("125")
        Tab0_Cbsbaud.Items.Add("250")
        Tab0_Cbsbaud.Items.Add("500")
        Tab0_Cfdbaud.Items.Add("1000")
        Tab0_Cfdbaud.Items.Add("2000")
        Tab0_Cfdbaud.Items.Add("5000")
    End Sub

    Private Sub Tab0_BLesPins_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles B_LesPins.Click
        Dim outstr, instr As String
        Dim ch0, ch1 As Integer

        'Auslesekommando ist 0x41 und dann noch zweites Byte (Wert irrelevant)
        outstr = Chr(&H41) + Chr(&H0)
        'Kommando senden
        UARTSend(outstr)
        'Auf Return warten
        instr = WaitForUARTReturn(2)
        'instr auswerten und Checkboxes entsprechend setzen
        ch0 = Asc(Mid(instr, 1, 1))
        ch1 = Asc(Mid(instr, 2, 1))
        If (ch1 And &H1) = &H1 Then Ch_Rselplus.Checked = True Else Ch_Rselplus.Checked = False
        If (ch1 And &H2) = &H2 Then Ch_Rledr.Checked = True Else Ch_Rledr.Checked = False
        If (ch1 And &H4) = &H4 Then Ch_Rledg.Checked = True Else Ch_Rledg.Checked = False
        If (ch1 And &H8) = &H8 Then Ch_Rbut.Checked = True Else Ch_Rbut.Checked = False
        If (ch1 And &H10) = &H10 Then Ch_Rsel1.Checked = True Else Ch_Rsel1.Checked = False
        If (ch1 And &H20) = &H20 Then Ch_Rsel2.Checked = True Else Ch_Rsel2.Checked = False
        If (ch1 And &H40) = &H40 Then Ch_Rsel4.Checked = True Else Ch_Rsel4.Checked = False
        If (ch1 And &H80) = &H80 Then Ch_Rsel8.Checked = True Else Ch_Rsel8.Checked = False
        If (ch0 And &H1) = &H1 Then Ch_Rsof.Checked = True Else Ch_Rsof.Checked = False
        If (ch0 And &H2) = &H2 Then Ch_Rint.Checked = True Else Ch_Rint.Checked = False
        If (ch0 And &H4) = &H4 Then Ch_Rint0.Checked = True Else Ch_Rint0.Checked = False
        If (ch0 And &H8) = &H8 Then Ch_Rint1.Checked = True Else Ch_Rint1.Checked = False
    End Sub

    Private Sub Tab0_BSetPins_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles B_SetPins.Click
        Dim bt1 As Integer = 0
        Dim outstr As String

        'Bits in Datenbyte entsprechend-Checkboxes setzen
        If Ch_Rselplus.Checked = True Then bt1 = bt1 Or &H1
        If Ch_Rledr.Checked = True Then bt1 = bt1 Or &H2
        If Ch_Rledg.Checked = True Then bt1 = bt1 Or &H4
        'Pinsetzen-Kommando ist 0x40 und zweites Byte mit entsprechendem Setting
        outstr = Chr(&H40) + Chr(bt1)
        UARTSend(outstr)
    End Sub

    Private Sub Tab0_Bread_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles B_read.Click
        Dim cmdstr, instr As String
        Dim bt, n, m, add, val As Integer

        T_read.Clear()
        If N_Add.Value >= &H400 And N_Add.Value < &HC00 Then    'Lesen aus dem RAM-Bereich des MCP ist Word-oriented (4 Byte)
            For n = 0 To N_Anz.Value - 1                    'Ein Kommando für 4 Bytes (N_Anz gibt die Anzahl der zu lesenen 4-Byte-Worte an)
                add = N_Add.Value + 4 * n                   'Adresse berechnen und eigenen cmdstr erzeugen
                val = add \ 256                             'Obere 4 Bit der 12-Bit-Adresse
                val = val Or &H30                           'Schreibkommando = 0x3?
                cmdstr = Chr(val)
                val = add Mod 256                           'Untere 8 Bit der 12-Bit Adresse
                cmdstr = cmdstr + Chr(val)
                UARTSend(cmdstr)
                instr = WaitForUARTReturn(4)                'Kommando liefert 4 Byte zurück
                For m = 0 To 3
                    bt = Asc(Mid(instr, m + 1, 1))
                    T_read.Text = T_read.Text + (add + m).ToString("X3") + ": " + bt.ToString("X2") + vbCrLf
                Next
            Next
        Else                                                    'Lesen aus dem SFR-Bereich des MCP ist Byte-oriented (1 Byte)
            For n = 0 To N_Anz.Value - 1                    'Byte für Byte ein eigenes Kommando
                add = N_Add.Value + n                       'Adresse berechnen und eigenen cmdstr erzeugen
                val = add \ 256                             'Obere 4 Bit der 12-Bit-Adresse
                val = val Or &H30                           'Schreibkommando = 0x3?
                cmdstr = Chr(val)
                val = add Mod 256                           'Untere 8 Bit der 12-Bit Adresse
                cmdstr = cmdstr + Chr(val)
                UARTSend(cmdstr)
                instr = WaitForUARTReturn(1)                'Kommando liefert 1 Byte zurück
                bt = Asc(Mid(instr, 1, 1))
                T_read.Text = T_read.Text + add.ToString("X3") + ": " + bt.ToString("X2") + vbCrLf
            Next
        End If
    End Sub

    Private Sub Tab0_Bwrite_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles B_write.Click
        Dim datstr, cmdstr, valtxt, valtxtmod As String
        Dim valtxtarray As String()
        Dim val, n, add As Integer

        'Aus den Eingaben der Textbox einen Datenstrom erstellen
        datstr = ""
        valtxtarray = (T_write.Text).Split(vbLf)
        For Each valtxt In valtxtarray
            valtxtmod = ""
            For n = 1 To valtxt.Length                                      'Alle Zeichen, die nicht Hex sind entfernen
                If Regex.IsMatch(Mid(valtxt, n, 1), "[a-fA-F0-9]") Then
                    valtxtmod = valtxtmod + Mid(valtxt, n, 1)
                End If
            Next
            If valtxtmod.Length > 0 Then
                If valtxtmod.Length > 2 Then                          'Bei mehr als zwei Stellen, nur die letzten 2 Stellen weiterverwenden
                    valtxtmod = Mid(valtxtmod, valtxtmod.Length - 1, 2)
                End If
                val = CInt("&H" + valtxtmod)
                datstr = datstr + Chr(val)
            End If
        Next
        If N_Add.Value >= &H400 And N_Add.Value < &HC00 Then    'Schreiben in den RAM-Bereich des MCP ist Word-oriented (4 Byte)
            n = 0
            While n < datstr.Length - 4                     'Jeweils immer 4 Byte pro Kommando
                add = N_Add.Value + n                       'Adresse berechnen und eigenen cmdstr erzeugen
                val = add \ 256                             'Obere 4 Bit der 12-Bit-Adresse
                val = val Or &H20                           'Schreibkommando = 0x2?
                cmdstr = Chr(val)
                val = add Mod 256                           'Untere 8 Bit der 12-Bit Adresse
                cmdstr = cmdstr + Chr(val)
                UARTSend(cmdstr)
                UARTSend(Mid(datstr, n + 1, 4))
                n = n + 4
            End While
        Else                                                    'Schreiben in den SFR-Bereich des MCP ist Byte-oriented (1 Byte)
            For n = 0 To datstr.Length - 1                  'Byte für Byte ein eigenes Kommando
                add = N_Add.Value + n                       'Adresse berechnen und eigenen cmdstr erzeugen
                val = add \ 256                             'Obere 4 Bit der 12-Bit-Adresse
                val = val Or &H20                           'Schreibkommando = 0x2?
                cmdstr = Chr(val)
                val = add Mod 256                           'Untere 8 Bit der 12-Bit Adresse
                cmdstr = cmdstr + Chr(val)
                UARTSend(cmdstr)
                UARTSend(Mid(datstr, n + 1, 1))
            Next
        End If
    End Sub

    Private Sub Tab0_Breset_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles B_reset.Click
        Dim outstr As String

        'Resetkommando ist 0x00 und dann noch zweites Byte (Wert irrelevant)
        outstr = Chr(&H0) + Chr(&H0)
        UARTSend(outstr)
    End Sub

    Private Sub Tab0_Breadbaud_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab0_Breadbaud.Click
        'Baudrate aus CAN-Controller auslesen
        Dim cmdstr, instr As String
        Dim startadd, add, n, val, bsbaud, fdbaud As Integer
        Dim bt(8) As Integer

        startadd = &H4                          'SFR 0x04-0x1B auslesen
        For n = 0 To 7                          'Byte für Byte ein eigenes Kommando
            add = startadd + n
            cmdstr = Chr(&H30) + Chr(add)       'Lesekommando = 0x3* + Adresse
            UARTSend(cmdstr)
            instr = WaitForUARTReturn(1)        'Kommando liefert 1 Byte zurück
            bt(n) = Asc(Mid(instr, 1, 1))
        Next
        bsbaud = 40000000 / (bt(2) + 2 + bt(1) + 1) / 1000       'Standard Baud-Berechnung - sehr vereinfacht, ohne baud-prescaler, etc. TSEG2 ist in bt(1), TSEG1 ist in bt(2)
        fdbaud = 40000000 / (bt(6) + 2 + bt(5) + 1) / 1000      'FD Baud-Berechnung - sehr vereinfacht, ohne baud-prescaler, etc. TSEG2 ist in bt(5), TSEG1 ist in bt(6)
        MsgBox("Base-Baudrate = " + bsbaud.ToString + "kbps" + vbNewLine + "FD-Baudrate = " + fdbaud.ToString + "kbps")
    End Sub

    Private Sub Tab0_Bsetbaud_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab0_Bsetbaud.Click
        'Baudrate im CAN-Controller setzen
        Dim cmdstr As String

        cmdstr = Chr(1)                 '0x01 = SetzeBaudrate-Kommando
        'Registerdefinitionen könnten auch berechnet werden, hier werden aber vordefinierte Werte genommen, siehe "MCP2517FD BaudRate Register Calculator.xlsx"
        If Tab0_Cbsbaud.Text = "125" Then
            cmdstr = cmdstr + Chr(63) + Chr(63) + Chr(254) + Chr(0)
        ElseIf Tab0_Cbsbaud.Text = "250" Then
            cmdstr = cmdstr + Chr(31) + Chr(31) + Chr(126) + Chr(0)
        ElseIf Tab0_Cbsbaud.Text = "500" Then
            cmdstr = cmdstr + Chr(15) + Chr(15) + Chr(62) + Chr(0)
        Else
            cmdstr = cmdstr + Chr(31) + Chr(31) + Chr(126) + Chr(0)
            MsgBox("Undefinierte Base-Baudrate. Wert wird auf 250kbps gesetzt.")
        End If
        If Tab0_Cfdbaud.Text = "1000" Then
            cmdstr = cmdstr + Chr(7) + Chr(7) + Chr(30) + Chr(0)
        ElseIf Tab0_Cfdbaud.Text = "2000" Then
            cmdstr = cmdstr + Chr(3) + Chr(3) + Chr(14) + Chr(0)
        ElseIf Tab0_Cfdbaud.Text = "5000" Then
            cmdstr = cmdstr + Chr(1) + Chr(1) + Chr(4) + Chr(0)
        Else
            cmdstr = cmdstr + Chr(7) + Chr(7) + Chr(30) + Chr(0)
            MsgBox("Undefinierte FD-Baudrate. Wert wird auf 1000kbps gesetzt.")
        End If
        UARTSend(cmdstr)
    End Sub

    Private Sub Tab0_Bfixbaud_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab0_Bfixbaud.Click
        'Baudrate im RaspberryPI in Startup-File schreiben (zum automatischen Laden bei Neustart)
        Dim cmdstr As String

        cmdstr = Chr(2) + Chr(0)        '0x02 = FixiereBaudrate-Kommando
        UARTSend(cmdstr)
    End Sub


    'Funktionen des TraceSID-Tab
    Private Sub Tab1_Init()
        Dim n As Integer
        Dim swidth, sheight, soben As Integer

        ReDim Tab1_TB(9)
        ReDim Tab1_PB(9)
        ReDim Tab1_LB(9)
        Tab1_LB(0) = New Label
        Tab1_LB(0).Text = "D" + vbNewLine + "L" + vbNewLine + "C"
        Tab1_LB(0).Left = 3
        Tab1_LB(0).Top = 72
        Tab1_LB(0).Width = 10
        Tab1_LB(0).Height = 40
        Me.TabSID.Controls.Add(Tab1_LB(0))
        Tab1_TB(0) = New TextBox
        Tab1_TB(0).Width = 39
        Tab1_TB(0).Height = 30
        Tab1_TB(0).Text = "00"
        Tab1_TB(0).Font = New Font("Microsoft Sans Serif", 15)
        Tab1_TB(0).Left = 24
        Tab1_TB(0).Top = 75
        Me.TabSID.Controls.Add(Tab1_TB(0))
        Tab1_PB(0) = New PictureBox
        Tab1_PB(0).Width = TabSID.Width - 80
        Tab1_PB(0).Height = 52
        Tab1_PB(0).Left = 69
        Tab1_PB(0).Top = 64
        Tab1_PB(0).BackColor = Color.White
        Tab1_PB(0).Image = New Bitmap(Tab1_PB(0).Width, Tab1_PB(0).Height)
        Me.TabSID.Controls.Add(Tab1_PB(0))

        'Verbelibende Resthohe und Breite für die restlichen 8 Sections berechnen (4 nebeneinander, 2 untereinander - Hoch genug, damit auch einzelne Bitänderungen erkannt werden können)
        Dim xanz As Integer = 4     '4 nebeneinander
        Dim yanz As Integer = 2     '2 untereinander
        swidth = TabSID.Width / xanz
        soben = Tab1_PB(0).Top + Tab1_PB(0).Height + 10
        sheight = (TabSID.Height - soben) / yanz
        'Verbleibende 8 Sections mit LAbel Textbox und Picturebox erstellen
        For n = 1 To 8
            Tab1_LB(n) = New Label
            Tab1_LB(n).Text = "B" + n.ToString()
            Tab1_LB(n).Width = 20
            Tab1_LB(n).Height = 40
            Tab1_LB(n).Left = 3 + ((n - 1) \ yanz) * swidth
            Tab1_LB(n).Top = soben + ((n - 1) Mod yanz) * sheight + 0.5 * sheight
            Me.TabSID.Controls.Add(Tab1_LB(n))

            Tab1_TB(n) = New TextBox
            Tab1_TB(n).Width = 39
            Tab1_TB(n).Height = 30
            Tab1_TB(n).Text = "00"
            Tab1_TB(n).Font = New Font("Microsoft Sans Serif", 15)
            Tab1_TB(n).Left = 24 + ((n - 1) \ yanz) * swidth
            Tab1_TB(n).Top = soben + ((n - 1) Mod yanz) * sheight + 0.5 * sheight - Tab1_TB(n).Height / 2
            Me.TabSID.Controls.Add(Tab1_TB(n))

            Tab1_PB(n) = New PictureBox
            Tab1_PB(n).Width = swidth - 80
            Tab1_PB(n).Height = sheight - 10
            Tab1_PB(n).Left = 69 + ((n - 1) \ yanz) * swidth
            Tab1_PB(n).Top = soben + ((n - 1) Mod yanz) * sheight
            Tab1_PB(n).BackColor = Color.White
            Tab1_PB(n).Image = New Bitmap(Tab1_PB(n).Width, Tab1_PB(n).Height)
            Me.TabSID.Controls.Add(Tab1_PB(n))
        Next
        Tab1_LCANID.Text = "CAN-ID: " + Tab1_SID.ToString("X3")
    End Sub

    Private Sub Tab1_BTrace_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab1_BTrace.Click
        Dim msg As String
        Dim SID, bt As Integer
        Dim Data, altData(9), n As Integer
        Dim neubild, altbild As Bitmap
        Dim CropRect As Rectangle
        Dim grp As Graphics
        Dim y, alty(9) As Integer
        Dim LinePen As Pen
        Dim ExitMsg As Boolean

        If Tab1_BTrace.Text = "Start Trace" Then
            msg = Chr(&H10) + Chr(&H0)          'Startkommando &H1*** an Gerät senden
            UARTSend(msg)
            Tab1_BTrace.Text = "Stop Trace"
            System.Windows.Forms.Application.DoEvents()
            LinePen = Pens.Blue
            For n = 0 To 8
                alty(n) = -1
                altData(n) = -1
            Next
            ExitMsg = False
            While Tab1_BTrace.Text = "Stop Trace" And Not ExitMsg
                msg = WaitForUARTReturn(16)         'Pro CAN-Message werden 16 Byte übertragen
                ExitMsg = True                     'Prüfe ob die Msg die Exit-Msg (16x 0x00) ist
                For n = 1 To 16
                    If Asc(Mid(msg, n, 1)) > 0 Then ExitMsg = False
                Next
                If Not ExitMsg Then                 'Wenn die Msg nicht die ExitMsg ist: Message verarbeiten
                    SID = Asc(Mid(msg, 1, 1))           'SID0-7 stecken im ersten Byte 
                    bt = Asc(Mid(msg, 2, 1))            'SID8-10 stecken in Bits 0-2 vom 2. Byte
                    SID = SID + (bt And &H7) * 256
                    bt = Asc(Mid(msg, 4, 1))            'SID11 steckt in Bit 6 von 4. Byte
                    If (bt And &H20) = &H20 Then SID = SID + 2048
                    If SID = Tab1_SID Then
                        If SetMark = True Then        'Wenn ein Marker gesetzt werden soll, soll Kurvenlinie Violet, ansonsten Blau gezeichnet werden. Zusätzlich werden alle anderen vertikalen Punkte der Stelle rot gezeichnet
                            LinePen = Pens.Violet       'Andere Farbe ist wichtig, sonst wäre kein Marker erkennbar, wenn er zufällig genau über einer Wertänderung von 0 auf 255 oder andersherum läge - Wertanzeioge geht von Anzeigen-Top bis Bottom ohne Rand
                        Else
                            LinePen = Pens.Blue
                        End If
                        bt = Asc(Mid(msg, 5, 1))        'DLC0-3 stecken in Bits0-3 vom 5. Byte
                        Data = (bt And &HF)              'DLC müsste hier eigentlich noch dekodiert werden, da bei CAN-FD hohere DLC-Werte nicht gleich der Anzahl Datenbytes sind (z.B. DLC=15 heißt 64 datenbytes)
                        If altData(0) = -1 Then altData(0) = Data 'beim erster Wert soll altData=Data sein
                        Tab1_TB(0).Text = Data.ToString
                        If Data <> altData(0) Then Tab1_TB(0).BackColor = Color.Red 'Change-Alarm setzen
                        'Im folgenden wird das bisherige Bild in der Picturebox um ein Pixel nach rechts geschoben und der neue Wert am linken Rand eingezeichnet
                        altbild = Tab1_PB(0).Image
                        neubild = New Bitmap(altbild.Width, altbild.Height)
                        CropRect = New Rectangle(1, 0, altbild.Width - 1, altbild.Height)
                        grp = Graphics.FromImage(neubild)
                        grp.DrawImage(altbild, New Rectangle(0, 0, CropRect.Width, CropRect.Height), CropRect, GraphicsUnit.Pixel)
                        y = (altbild.Height - 2) - Data / 15 * (altbild.Height - 2)
                        If alty(0) = -1 Then alty(0) = y 'beim erster Wert soll alty=y sein
                        If SetMark = True Then
                            grp.DrawLine(Pens.Red, altbild.Width - 1, 0, altbild.Width - 1, altbild.Height)         'Vertikale rote Marker-Linie einzeichnen
                        End If
                        If alty(0) = y Then             'Neuen Punkt (2 Pixel hoch) zeichnen, jweils unterscheidend ob Wert von "unten" nach oben oder anders herunm geht
                            grp.DrawLine(LinePen, altbild.Width - 1, y, altbild.Width - 1, y + 1)         'Höhe gleichbleiben
                        ElseIf alty(0) > y Then
                            grp.DrawLine(LinePen, altbild.Width - 1, alty(0), altbild.Width - 1, y)       'Linie Von unten nach oben
                        Else
                            grp.DrawLine(LinePen, altbild.Width - 1, alty(0) + 1, altbild.Width - 1, y + 1)   'Linie von oben nach unten
                        End If
                        Tab1_PB(0).Image = neubild
                        alty(0) = y
                        altData(0) = Data
                        '8 Datenbytes auswerten, Stecken in Bytes 9-16 der Message
                        For n = 1 To 8
                            Data = Asc(Mid(msg, 8 + n, 1))
                            If altData(n) = -1 Then altData(n) = Data 'beim erster Wert soll altData=Data sein
                            Tab1_TB(n).Text = Data.ToString("X2")
                            If Data <> altData(n) Then Tab1_TB(n).BackColor = Color.Red 'Change-Alarm setzen
                            'Im folgenden wird das bisherige Bild in der Picturebox um ein Pixel nach rechts geschoben und der neue Wert am linken Rand eingezeichnet
                            altbild = Tab1_PB(n).Image
                            neubild = New Bitmap(altbild.Width, altbild.Height)
                            CropRect = New Rectangle(1, 0, altbild.Width - 1, altbild.Height)
                            grp = Graphics.FromImage(neubild)
                            grp.DrawImage(altbild, New Rectangle(0, 0, CropRect.Width, CropRect.Height), CropRect, GraphicsUnit.Pixel)
                            y = (altbild.Height - 2) - Data / 255 * (altbild.Height - 2)
                            If alty(n) = -1 Then alty(n) = y 'beim erster Wert soll alty=y sein
                            If SetMark = True Then
                                grp.DrawLine(Pens.Red, altbild.Width - 1, 0, altbild.Width - 1, altbild.Height)         'Vertikale rote Marker-Linie einzeichnen
                            End If
                            If alty(n) = y Then             'Neuen Punkt (2 Pixel hoch) zeichnen, jweils unterscheidend ob Wert von "unten" nach oben oder anders herunm geht
                                grp.DrawLine(LinePen, altbild.Width - 1, y, altbild.Width - 1, y + 1)         'Höhe gleichbleiben
                            ElseIf alty(n) > y Then
                                grp.DrawLine(LinePen, altbild.Width - 1, alty(n), altbild.Width - 1, y)       'Linie Von unten nach oben
                            Else
                                grp.DrawLine(LinePen, altbild.Width - 1, alty(n) + 1, altbild.Width - 1, y + 1)   'Linie von oben nach unten
                            End If
                            Tab1_PB(n).Image = neubild
                            alty(n) = y
                            altData(n) = Data
                        Next
                        SetMark = False       'Markierungsflag zurücksetzen
                    Else
                        'MsgBox("Irgendwie wurde eine nicht passende SID empfangen: " + SID.ToString("X3"))
                        T_RX.Text = T_RX.Text + "Falsche ID!" + SID.ToString("X3") + vbNewLine
                    End If
                Else                            'Es wurde eine Exit-Msg empfangen (d.h. am Gerät wurde die Taste gedrückt und damit der Tracing-Modus beendet: d.h. Schleife und Sub B_Trace_Click werden beendet
                    Tab1_BTrace.Text = "Start Trace"
                    System.Windows.Forms.Application.DoEvents()
                    MsgBox("Am Gerät wurde die Taste gedrückt - Tracing-Modus wurde beendet")
                End If
            End While
        Else
            msg = Chr(&H20) + Chr(&H0)          'Stopkommando &H2*** an Gerät senden
            UARTSend(msg)
            Tab1_BTrace.Text = "Start Trace"
            System.Windows.Forms.Application.DoEvents()
            Threading.Thread.Sleep(100)         '0,1s warten um sicherzustellen, dass Kommando angekommen ist und Gerät auch nicht mehr weitersendet
            UART_ClearInput()                   'Evtl zwischenzeitlich noch angekommene und nicht mehr verarbeitete Daten verwerfen
        End If
    End Sub

    Private Sub Tab1_BIDchange_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab1_BIDchange.Click
        Dim id, val As Integer
        Dim cmdstr As String

        id = Tab1_NID.Value                       'Adresse berechnen und cmdstr erzeugen
        val = id \ 256                             'Obere 4 Bit der 12-Bit-ID
        cmdstr = Chr(val)                           'Kommando-Bits (obere 4 Bits) sind damit automatisch 0x0
        val = id Mod 256                           'Untere 8 Bit der 12-Bit ID
        cmdstr = cmdstr + Chr(val)
        UARTSend(cmdstr)
        Tab1_SID = id
        Tab1_LCANID.Text = "CAN-ID: " + id.ToString("X3")
        Tab2_LCANID.Text = "CAN-ID: " + id.ToString("X3")
    End Sub

    Private Sub Tab1_BClear_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab1_BClear.Click
        Dim n As Integer

        For n = 0 To 8
            Tab1_PB(n).Image = New Bitmap(Tab1_PB(n).Width, Tab1_PB(n).Height)
        Next
    End Sub

    Private Sub Tab1_BAlarmReset_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab1_BAlarmReset.Click
        'Resettet alle Änderungsalarme
        Dim n As Integer

        For n = 0 To 8
            Tab1_TB(n).BackColor = Color.White
        Next

    End Sub

    Private Sub Tab1_BMarkerSet_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab1_BMarkerSet.Click
        'Setzt ein Flag, das signalisiert, das in den Grafiken ein Marker gesetzt wird

        SetMark = True        'Beim nächsten Zeichnen wird Marker eingezeichnet und Flag zurückgesetzt
    End Sub


    'Funktionen des TraceByteSeq-Tab
    Private Sub Tab2_Init()
        Tab2_LCANID.Text = "CAN-ID: " + Tab1_SID.ToString("X3")
        Tab2_PB.BackColor = Color.White
        Tab2_PB.Image = New Bitmap(Tab2_PB.Width, Tab2_PB.Height)
    End Sub

    Private Sub Tab2_BTrace_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab2_BTrace.Click
        Dim msg As String
        Dim SID, byt, bit As Integer
        Dim Data, altData, n As Integer
        Dim neubild, altbild As Bitmap
        Dim CropRect As Rectangle
        Dim grp As Graphics
        Dim y, alty As Integer
        Dim LinePen As Pen
        Dim ExitMsg As Boolean
        Dim bits() As String
        Dim bytenr, bitnr As Integer

        If Tab2_BTrace.Text = "Start Trace" Then
            msg = Chr(&H10) + Chr(&H0)          'Startkommando &H1*** an Gerät senden
            UARTSend(msg)
            Tab2_BTrace.Text = "Stop Trace"
            System.Windows.Forms.Application.DoEvents()
            LinePen = Pens.Blue
            For n = 0 To 8
                alty = -1
                altData = -1
            Next
            ExitMsg = False
            While Tab2_BTrace.Text = "Stop Trace" And Not ExitMsg
                msg = WaitForUARTReturn(16)         'Pro CAN-Message werden 16 Byte übertragen
                ExitMsg = True                     'Prüfe ob die Msg die Exit-Msg (16x 0x00) ist
                For n = 1 To 16
                    If Asc(Mid(msg, n, 1)) > 0 Then ExitMsg = False
                Next
                If Not ExitMsg Then                 'Wenn die Msg nicht die ExitMsg ist: Message verarbeiten
                    SID = Asc(Mid(msg, 1, 1))           'SID0-7 stecken im ersten Byte 
                    byt = Asc(Mid(msg, 2, 1))            'SID8-10 stecken in Bits 0-2 vom 2. Byte
                    SID = SID + (byt And &H7) * 256
                    byt = Asc(Mid(msg, 4, 1))            'SID11 steckt in Bit 6 von 4. Byte
                    If (byt And &H20) = &H20 Then SID = SID + 2048
                    If SID = Tab1_SID Then
                        If SetMark = True Then        'Wenn ein Marker gesetzt werden soll, soll Kurvenlinie Violet, ansonsten Blau gezeichnet werden. Zusätzlich werden alle anderen vertikalen Punkte der Stelle rot gezeichnet
                            LinePen = Pens.Violet       'Andere Farbe ist wichtig, sonst wäre kein Marker erkennbar, wenn er zufällig genau über einer Wertänderung von 0 auf 255 oder andersherum läge - Wertanzeioge geht von Anzeigen-Top bis Bottom ohne Rand
                        Else
                            LinePen = Pens.Blue
                        End If
                        'gewünschte Bitsequenz ausarbeiten und gewünschten Datenwert berechnen
                        bits = Tab2_TBitSequenz.Text.Split("_")
                        Data = 0
                        For n = 1 To bits.Length
                            bytenr = Val(Mid(bits(n - 1), 1, 1))
                            bitnr = Val(Mid(bits(n - 1), 3, 1))
                            byt = Asc(Mid(msg, 8 + bytenr, 1))       'Gewünschtes Byte
                            bit = (byt And (2 ^ (bitnr - 1))) / (2 ^ (bitnr - 1)) 'gewünschtes Bit im gewünschten Byte (0 oder 1)
                            Data = Data + bit * (2 ^ (bits.Length - n))
                        Next
                        If SetMark = True Then
                            Tab2_TProt.Text = "*h" + Data.ToString("X") + " : d" + Data.ToString + vbNewLine + Tab2_TProt.Text
                        Else
                            Tab2_TProt.Text = "h" + Data.ToString("X") + " : d" + Data.ToString + vbNewLine + Tab2_TProt.Text
                        End If
                        If altData = -1 Then altData = Data 'beim erster Wert soll altData=Data sein
                        Tab2_TWert.Text = Data.ToString
                        If Data <> altData Then Tab2_TWert.BackColor = Color.Red 'Change-Alarm setzen
                        'Im folgenden wird das bisherige Bild in der Picturebox um ein Pixel nach rechts geschoben und der neue Wert am linken Rand eingezeichnet
                        altbild = Tab2_PB.Image
                        neubild = New Bitmap(altbild.Width, altbild.Height)
                        CropRect = New Rectangle(1, 0, altbild.Width - 1, altbild.Height)
                        grp = Graphics.FromImage(neubild)
                        grp.DrawImage(altbild, New Rectangle(0, 0, CropRect.Width, CropRect.Height), CropRect, GraphicsUnit.Pixel)
                        y = (altbild.Height - 2) - Data / ((2 ^ bits.Length) - 1) * (altbild.Height - 2)
                        If alty = -1 Then alty = y 'beim erster Wert soll alty=y sein
                        If SetMark = True Then
                            grp.DrawLine(Pens.Red, altbild.Width - 1, 0, altbild.Width - 1, altbild.Height)         'Vertikale rote Marker-Linie einzeichnen
                        End If
                        If alty = y Then             'Neuen Punkt (2 Pixel hoch) zeichnen, jweils unterscheidend ob Wert von "unten" nach oben oder anders herunm geht
                            grp.DrawLine(LinePen, altbild.Width - 1, y, altbild.Width - 1, y + 1)         'Höhe gleichbleiben
                        ElseIf alty > y Then
                            grp.DrawLine(LinePen, altbild.Width - 1, alty, altbild.Width - 1, y)       'Linie Von unten nach oben
                        Else
                            grp.DrawLine(LinePen, altbild.Width - 1, alty + 1, altbild.Width - 1, y + 1)   'Linie von oben nach unten
                        End If
                        Tab2_PB.Image = neubild
                        alty = y
                        altData = Data
                        SetMark = False       'Markierungsflag zurücksetzen
                    Else
                        'MsgBox("Irgendwie wurde eine nicht passende SID empfangen: " + SID.ToString("X3"))
                        T_RX.Text = T_RX.Text + "Falsche ID!" + SID.ToString("X3") + vbNewLine
                    End If
                Else                            'Es wurde eine Exit-Msg empfangen (d.h. am Gerät wurde die Taste gedrückt und damit der Tracing-Modus beendet: d.h. Schleife und Sub B_Trace_Click werden beendet
                    Tab2_BTrace.Text = "Start Trace"
                    System.Windows.Forms.Application.DoEvents()
                    MsgBox("Am Gerät wurde die Taste gedrückt - Tracing-Modus wurde beendet")
                End If
            End While
        Else
            msg = Chr(&H20) + Chr(&H0)          'Stopkommando &H2*** an Gerät senden
            UARTSend(msg)
            Tab2_BTrace.Text = "Start Trace"
            System.Windows.Forms.Application.DoEvents()
            Threading.Thread.Sleep(100)         '0,1s warten um sicherzustellen, dass Kommando angekommen ist und Gerät auch nicht mehr weitersendet
            UART_ClearInput()                   'Evtl zwischenzeitlich noch angekommene und nicht mehr verarbeitete Daten verwerfen
        End If
    End Sub

    Private Sub Tab2_BIDchange_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab2_BIDchange.Click
        Dim id, val As Integer
        Dim cmdstr As String

        id = Tab2_NID.Value                       'Adresse berechnen und cmdstr erzeugen
        val = id \ 256                             'Obere 4 Bit der 12-Bit-ID
        cmdstr = Chr(val)                           'Kommando-Bits (obere 4 Bits) sind damit automatisch 0x0
        val = id Mod 256                           'Untere 8 Bit der 12-Bit ID
        cmdstr = cmdstr + Chr(val)
        UARTSend(cmdstr)
        Tab1_SID = id
        Tab1_LCANID.Text = "CAN-ID: " + id.ToString("X3")
        Tab2_LCANID.Text = "CAN-ID: " + id.ToString("X3")
    End Sub

    Private Sub Tab2_BClear_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab2_BClear.Click
        Tab2_PB.Image = New Bitmap(Tab2_PB.Width, Tab2_PB.Height)
        Tab2_TProt.Clear()
    End Sub

    Private Sub Tab2_BAlarmReset_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab2_BAlarmReset.Click
        'Resettet Änderungsalarm

        Tab2_TWert.BackColor = Color.White
    End Sub

    Private Sub Tab2_BMarkerSet_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab2_BMarkerSet.Click
        'Setzt ein Flag, das signalisiert, das in der Grafik ein Marker gesetzt wird

        SetMark = True        'Beim nächsten Zeichnen wird Marker eingezeichnet und Flag zurückgesetzt
    End Sub


    'Funktionen des RecordAll-Tab
    Private Sub Tab3_BStartStop_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab3_BStartStop.Click
        Dim msg, SIDt As String
        Dim ExitMsg, Found As Boolean
        Dim n, byt1, byt2, byt3 As Integer
        Dim SID, FDF, RTR, IDE, BRS, DLC As Integer

        If Tab3_BStartStop.Text = "Start Protokoll" Then
            Tab3_BStartStop.Text = "Stop Protokoll"
            System.Windows.Forms.Application.DoEvents()
            ExitMsg = False
            While Tab3_BStartStop.Text = "Stop Protokoll" And Not ExitMsg
                msg = WaitForUARTReturn(3)         'Pro CAN-Message werden 3 Byte übertragen
                ExitMsg = True                     'Prüfe ob die Msg die Exit-Msg (3x 0x00) ist
                For n = 1 To 3
                    If Asc(Mid(msg, n, 1)) > 0 Then ExitMsg = False
                Next
                If Not ExitMsg Then                 'Wenn die Msg nicht die ExitMsg ist: Message verarbeiten
                    byt1 = Asc(Mid(msg, 1, 1))      '1.Summary-Byte - Wenn Synchronisiert: 1, BRS, RTR, IDE, DLC(3:0)
                    byt2 = Asc(Mid(msg, 2, 1))      '2.Summary-Byte - Wenn Synchronisiert: 0, FDF, 0, 0, SID(11:8)
                    byt3 = Asc(Mid(msg, 3, 1))      '3.Summary-Byte - Wenn Synchronisiert: SID(7:0)
                    If (((byt1 And &H80) = &H80) And ((byt2 And &H80) = &H0)) Then    'Ist die Übertragung noch synchronisiert (oberstes Bit von Byte1=1 und oberstes Bit von Byte2=0?
                        FDF = (byt2 And &H40) / &H40        'FDF-Bit wird in Bit 6 des Byte2 übertragen
                        SID = (byt2 And &HF) * 256          'Bits 8-11 der SID werden in Bits 0-3 des Byte2 übertragen
                        SID = SID + byt3                    'Bits 0-7 der SID werden in Bits 0-7 des Byte3 übertragen
                        SIDt = SID.ToString("X3")
                        BRS = (byt1 And &H40) / &H40        'BRS-Bit wird in Bit 6 des Byte1 übertragen
                        RTR = (byt1 And &H20) / &H20        'RTR-Bit wird in Bit 5 des Byte1 übertragen
                        IDE = (byt1 And &H10) / &H10        'IDE-Bit wird in Bit 4 des Byte1 übertragen
                        DLC = (byt1 And &HF)                'DLC wird in Bits 0-4 des Byte1 übertragen - Hier müsste eigentlich noch Auswertung rein (in CAN-FD bedeutet DLC=15 z.B. 64 datenbytes)
                        Found = False
                        For n = 0 To Tab3_List.Rows.Count - 1       'Suchen ob schon ein Eintrag mit diesen Daten vorhanden ist
                            If Tab3_List.Item("SID", n).Value = SIDt And Tab3_List.Item("FDF", n).Value = FDF And Tab3_List.Item("BRS", n).Value = BRS And Tab3_List.Item("RTR", n).Value = RTR And Tab3_List.Item("IDE", n).Value = IDE And Tab3_List.Item("DLC", n).Value = DLC Then
                                Tab3_List.Item("CNT", n).Value = Tab3_List.Item("CNT", n).Value + 1         'Eintrag wurde gefunden: Count erhöhen und Schleife verlassen
                                Found = True
                                n = Tab3_List.Rows.Count - 1
                            End If
                        Next
                        If Found = False Then               'Es wurde kein Eintrag gefunden, der genau dem eben empfangenen entspricht
                            Tab3_List.Rows.Add(1, SID.ToString("X3"), FDF, BRS, RTR, IDE, DLC)     'Neuen Eintrag erstellen
                        End If
                    Else                                    'Übertragung ist nicht synchronisiert
                        msg = WaitForUARTReturn(1)         'Neu-Synchronisierung versuchen, indem ein empfangenes Byte verworfen wird
                    End If
                Else                            'Es wurde eine Exit-Msg empfangen (d.h. am Gerät wurde die Taste gedrückt und damit der Tracing-Modus beendet: d.h. Schleife und Sub B_Trace_Click werden beendet
                    Tab3_BStartStop.Text = "Start Protokoll"
                    System.Windows.Forms.Application.DoEvents()
                    MsgBox("Am Gerät wurde die Taste gedrückt - Record-Modus wurde beendet")
                End If
            End While
        Else
            Tab3_BStartStop.Text = "Start Protokoll"
            System.Windows.Forms.Application.DoEvents()
        End If
    End Sub

    Private Sub Tab3_BClear_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab3_BClear.Click
        Tab3_List.Rows.Clear()
    End Sub


    'Funktionen des SendMessage-Tab
    Private Sub Tab4_Init()
        '64 (8x8) Datenfelder mit Label generieren
        Dim n, swidth, soben, slinks, sheight As Integer

        ReDim Tab4_LD(64)
        ReDim Tab4_ND(64)
        Dim xanz As Integer = 8     '8 nebeneinander
        Dim yanz As Integer = 8     '8 untereinander
        soben = GroupBox3.Top + 6
        slinks = Tab4_Nid.Right + 10
        swidth = (TabTmit.Width - slinks) / 2 / (xanz - 1)
        sheight = (Tab4_Ndlc.Bottom - soben) / (yanz - 1)
        For n = 0 To 63
            Tab4_LD(n) = New Label
            Tab4_LD(n).Text = "D" + (n + 1).ToString() + ":"
            Tab4_LD(n).Width = 28
            Tab4_LD(n).Height = 40
            Tab4_LD(n).Left = slinks + (n Mod yanz) * swidth
            Tab4_LD(n).Top = soben + (n \ yanz) * sheight
            Tab4_LD(n).Visible = False
            Me.TabTmit.Controls.Add(Tab4_LD(n))

            Tab4_ND(n) = New NumericUpDown
            Tab4_ND(n).Width = 39
            Tab4_ND(n).Height = 30
            Tab4_ND(n).Left = Tab4_LD(n).Left + Tab4_LD(n).Width + 3
            Tab4_ND(n).Top = Tab4_LD(n).Top - 3
            Tab4_ND(n).Visible = False
            Tab4_ND(n).Minimum = 0
            Tab4_ND(n).Maximum = 255
            Tab4_ND(n).Value = 0
            Tab4_ND(n).Hexadecimal = True
            Tab4_ND(n).TextAlign = HorizontalAlignment.Right
            Me.TabTmit.Controls.Add(Tab4_ND(n))
        Next
    End Sub

    Private Sub Tab4_BTmit_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab4_BTmit.Click
        'Sendet definierte Message - Frame zusammenstellen und an Controller übertragen
        Dim header(8), n, dlc As Integer
        Dim txstr, rxstr As String


        '8 Byte Header zusammenstellen und an Gerät senden
        header(0) = Tab4_Nid.Value And &HFF         'Header0(0..7) beinhaltet SID(0..7)
        header(1) = (Tab4_Nid.Value >> 8) And &H7   'Header1(0..2) beinhaltet SID(8..10), (3..7) wäre EID(0..4) - wird aktuell noch nicht unterstützt
        header(2) = 0                               'Header2(0..7) wäre EID(5..12) - wird aktuell noch nicht unterstützt
        header(3) = (Tab4_Nid.Value >> 6) And &H20  'Header3(5) beinhaltet SID(11), (0..4) wäre EID(13..17) - wird aktuell noch nicht unterstützt
        header(4) = Tab4_Ndlc.Value                 'Header4(0..3) beinhaltet DLC(0..4)
        If Tab4_RExtID.Checked = True Then header(4) = header(4) Or &H10 'Header4(4) beinhaltet IDE-Bit (Extended ID)
        If Tab4_RRemFrm.Checked = True Then header(4) = header(4) Or &H20 'Header4(5) beinhaltet RTR-Bit (Remote Frame)
        If Tab4_RCANFD.Checked = True Then
            header(4) = header(4) Or &H80           'Header4(7) beinhaltet FDF-Bit (CAN-FD)
            header(4) = header(4) Or &H40           'Header4(6) beinhaltet BRS-Bit (Bitrate-Switch) - bei CAN-FD wird davon ausgegangen, dass geswitcht werden soll
        End If
        header(5) = 0                               'Header(5) bis Header(7) sind im MCP2517 nicht besetzt (wäre SEQ bei MCP2518)
        header(6) = 0
        header(7) = 0
        txstr = ""
        For n = 0 To 7
            txstr = txstr + Chr(header(n))
        Next
        UARTSend(txstr)

        'Warte auf DLC-Request vom Gerät
        rxstr = WaitForUARTReturn(1)            'DLC-Request ist 1 Byte lang
        dlc = Asc(Mid(rxstr, 1, 1))
        If dlc <= 64 Then                       'DLC>64 kann eigentlich nicht sein - trotzdem, um fehler zu vermeiden:
            'Gerät erwartet jetzt die Übermittlung von dlc Daten-Bytes (könnte man noch gegen-checken gegen DLC-Value, der im Header verschickt wurde...)
            'Daten-Übertragung vorbereiten und dann senden
            txstr = ""
            For n = 0 To dlc - 1
                txstr = txstr + Chr(Tab4_ND(n).Value)
            Next
            UARTSend(txstr)
        End If
    End Sub

    Private Sub Tab4_RCAN20_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab4_RCAN20.CheckedChanged
        'CAN2.0 oder CAN-FD Frame

        If initialized Then
            If Tab4_RCAN20.Checked = True Then
                ShowDataFields(Tab4_Ndlc.Value)
            Else
                ShowDataFields(Tab4_Ndlc.Value)
            End If
        End If
    End Sub

    Private Sub Tab4_RBaseID_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab4_RBaseID.CheckedChanged
        'Base-ID oder Extended ID

        If initialized Then
            If Tab4_RBaseID.Checked = True Then
                Tab4_Nid.Maximum = 4095         'Eigentlich nur 11 Bit
            Else
                Tab4_Nid.Maximum = 536870911    '29 Bit extended ID
            End If
        End If
    End Sub

    Private Sub Tab4_RDatFrm_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab4_RDatFrm.CheckedChanged
        'Data Frame oder remote Frame

        If initialized Then
            If Tab4_RDatFrm.Checked = True Then
                If Tab4_RCAN20.Checked = True Then
                    ShowDataFields(Tab4_Ndlc.Value)
                Else
                    ShowDataFields(Tab4_Ndlc.Value)
                End If
            Else
                ShowDataFields(0)
            End If
        End If
    End Sub

    Private Sub Tab4_Ndlc_ValueChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Tab4_Ndlc.ValueChanged

        If initialized Then
            If Tab4_RCAN20.Checked = True Then
                ShowDataFields(Tab4_Ndlc.Value)
            Else
                ShowDataFields(Tab4_Ndlc.Value)
            End If
        End If
    End Sub

    Private Sub ShowDataFields(ByVal anz As Integer)
        'Setzt erforderliche Anzahl Datenfelder visible, den Rest invisible
        Dim n As Integer

        'Erforderliche Anzahl ermitteln - je nach CAN2.0 oder CAN-FD
        If Tab4_RCAN20.Checked = True Then      'CAN2.0
            If anz > 8 Then anz = 8
        Else                                    'CAN-FD
            If anz = 9 Then
                anz = 12
            ElseIf anz = 10 Then
                anz = 16
            ElseIf anz = 11 Then
                anz = 20
            ElseIf anz = 12 Then
                anz = 24
            ElseIf anz = 13 Then
                anz = 32
            ElseIf anz = 14 Then
                anz = 48
            ElseIf anz = 15 Then
                anz = 64
            End If
        End If
        'Entsprechend erste Anzahl Datenfelder visible, den Rest invisible
        If anz > 0 Then
            For n = 0 To anz - 1
                Tab4_ND(n).Visible = True
                Tab4_LD(n).Visible = True
            Next
        End If
        For n = anz To 63
            Tab4_ND(n).Visible = False
            Tab4_LD(n).Visible = False
        Next
    End Sub

End Class
